use std::marker::PhantomData;
use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer, type_key,
        clone,
        array_set, array_at,
    },
};

pub struct Array<const N :usize, T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<const N :usize, T: SzunType> Array<N,T> {
    pub fn new() -> Self
    {
        Self { address:unsafe {acquire(Self::type_id())}, _data:PhantomData }
    }

    pub fn set<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {array_set(self.address.clone(), index, data.__to_ref())};
        }
    }

    pub fn get(&self, index:usize) -> T
    {
        let refr = unsafe {array_at(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("array index out of bounds"); }
    }

    pub fn len(&self) -> usize
    {
        N
    }
}
impl<const N :usize, T: SzunType> SzunType for Array<N,T> {
    fn type_id() -> usize {
        let inner = unsafe {type_outer(T::type_id(), N)};
        unsafe {type_outer(inner, types::TYPE_ARRAY)}
    }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<R: SzunType>(&self, _:&R) -> bool {
        R::type_id() == T::type_id() || unsafe {type_key(T::type_id())} == types::TYPE_VARYING
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr, _data:PhantomData } }
}

impl<const N :usize, T: SzunType> Clone for Array<N,T> {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())}, _data:PhantomData }
    }
}

impl<const N :usize, T: SzunType> TryFrom<Ref> for Array<N,T> {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr, _data:PhantomData })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<const N :usize, T: SzunType> Into<Ref> for Array<N,T> { fn into(self) -> Ref { self.address.clone() } }

impl<const N :usize, T: SzunType> Drop for Array<N,T> {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
