use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer,
        clone,
        bool_set, bool_get,
    },
};

pub struct Boolean {
    address:Ref,
}
impl Boolean {
    pub fn new() -> Self
    {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:bool) -> Self
    {
        self.set(data);
        self
    }

    pub fn set(&mut self, data:bool)
    {
        unsafe {bool_set(self.address.clone(), data)}
    }
    
    pub fn get(&self) -> bool
    {
        unsafe {bool_get(self.address.clone())}
    }
}
impl SzunType for Boolean {
    fn type_id() -> usize { unsafe {type_outer(0, types::TYPE_BOOLEAN)} }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }
    
    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn boolean(data:bool) -> Boolean
{
    Boolean::from(data)
}

impl Clone for Boolean {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

impl std::ops::Not for Boolean {
    type Output = Self;
    fn not(self) -> Self::Output {
        Self::from(!self.get())
    }
}

impl TryFrom<Ref> for Boolean {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Ref> for Boolean { fn into(self) -> Ref { self.address.clone() } }

impl From<bool> for Boolean { fn from(value: bool) -> Self { Self::new().with(value) } }
impl Into<bool> for Boolean { fn into(self) -> bool { self.get() } }

impl Drop for Boolean {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
