use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer,
        clone,
        integer_set, integer_get,
    },
};

pub struct Integer {
    address:Ref,
}
impl Integer {
    pub fn new() -> Self {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with<T: num::Num + num::NumCast>(mut self, data:T) -> Self
    {
        self.set(data);
        self
    }

    pub fn set<T: num::Num + num::NumCast>(&mut self, data:T) {
        match num::NumCast::from(data) {
            Some(data) => unsafe {integer_set(self.address.clone(), data)}
            None => { }
        }
    }
    
    pub fn get<T: num::Num + num::NumCast>(&self) -> T {
        num::NumCast::from(unsafe {integer_get(self.address.clone())}).unwrap()
    }
}
impl SzunType for Integer {
    fn type_id() -> usize { unsafe {type_outer(0, types::TYPE_INTEGER)} }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn integer<T: num::Num + num::NumCast>(data:T) -> Integer
{
    Integer::new().with(data)
}

impl Clone for Integer {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

impl std::ops::Neg for Integer {
    type Output = Self;
    fn neg(self) -> Self::Output {
        Self::new().with(-self.get::<i64>())
    }
}

crate::util::macros::impl_ops_numeric_all!(Integer, i64);

impl TryFrom<Ref> for Integer {
    type Error = SzunError;
    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Ref> for Integer { fn into(self) -> Ref { self.address.clone() } }

impl Drop for Integer {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
