use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer,
        clone,
        natural_set, natural_get,
    },
};

pub struct Natural {
    address:Ref,
}
impl Natural {
    pub fn new() -> Self {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with<T: num::Num + num::NumCast>(mut self, data:T) -> Self {
        self.set(data);
        self
    }

    pub fn set<T: num::Num + num::NumCast>(&mut self, data:T) {
        match num::NumCast::from(data) {
            Some(data) => unsafe {natural_set(self.address.clone(), data)}
            None => { }
        }
    }
    
    pub fn get<T: num::Num + num::NumCast>(&self) -> T {
        num::NumCast::from(unsafe {natural_get(self.address.clone())}).unwrap()
    }
}
impl SzunType for Natural {
    fn type_id() -> usize { unsafe {type_outer(0, types::TYPE_NATURAL)} }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn natural<T: num::Num + num::NumCast>(data:T) -> Natural
{
    Natural::new().with(data)
}

impl Clone for Natural {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

crate::util::macros::impl_ops_numeric_all!(Natural, u64);

impl From<Ref> for Natural {
    fn from(refr: Ref) -> Self {
        if refr.class == Self::type_id() {
            Self { address:refr }
        } else { panic!("mismatched szun types"); }
    }
}
impl Into<Ref> for Natural { fn into(self) -> Ref { self.address.clone() } }

impl Drop for Natural {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
